#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          31-07-2019
//      Created for Project :   FLAME
"""

#
# This script could be used to generate the list of requirements (a.k.a. python requirements.txt) in the correct dependency order with fixed versions for all packages. An example would be
# if CLMC depends on package1 and package2 which both depend on package3 which depends on package4, the requirements order would be package4, package3, package2, package1.
# The script expects a dependency tree as input which must be produced by the python pipdeptree utility library.
#

from collections import defaultdict


def generate_dependency_mapping(dependency_tree):
    """
    Traverses the dependency tree generated by the pipdeptree utility and groups each dependency to a depth level

    :param dependency_tree: the dependency tree generated from the 'pipdeptree --reverse' command
    :return: a dictionary mapping each dependency to a depth level (depth 0 dependencies must be installed first, depth 1 after depth 0, etc.)
            e.g. {0: {package1, package2}, 1: {package3}, 2: {package4, package5, package6}}
    """

    dependency_depths = defaultdict(set)
    found = {}

    # traverse each line
    for line in dependency_tree.split("\n"):
        if line == "":
            continue  # ignore empty lines

        # each line represents a dependency package, the number of spaces in the beginning determines the depth of this dependency
        spaces_count = len(line) - len(line.lstrip(" "))

        # spaces are multiples of 2
        depth = int(spaces_count / 2)

        package_name = line.lstrip(" ")
        if package_name == "":
            continue  # ignore lines that consisted of spaces only

        if package_name.startswith("-"):
            # pipdeptree will output a dash and a space in front of a dependency if the depth is greater than 0
            package_name = package_name[2:]

            # pipdeptree will also include information on which package requires this dependency, strip that info as well
            package_name_end_index = package_name.index(" [")
            package_name = package_name[: package_name_end_index]

        if package_name.startswith("pipdeptree==") or package_name.startswith("pip==") or package_name.startswith("clmcservice=="):
            continue  # ignore the clmcservice, pipdeptree and pip dependencies (pip is installed anyway before installing the generated list of requirements)

        # if this is a new package, just add it to the appropriate depth level
        if package_name not in found:
            dependency_depths[depth].add(package_name)
            found[package_name] = depth
        # if package name already seen, and the current depth is greater than the previous depth, add it to the current depth and remove it from previous depth packages
        else:
            previous_depth = found[package_name]
            if depth > previous_depth:
                dependency_depths[previous_depth].remove(package_name)
                dependency_depths[depth].add(package_name)
                found[package_name] = depth

    return dependency_depths


def generate_dependency_requirements(dependency_mapping):
    """
    Generates a list of requirements in the order that they have to be installed.

    :param dependency_mapping: a mapping from a depth level to a set of packages.

    :return: a string each line of which represents a package to install
    """

    # find the maximum depth level in the list of dependencies
    max_depth = max(dependency_mapping.keys())

    # a list to save the output for each depth
    outputs = []

    # go through each depth and generate the output for the dependencies at this depth
    for depth in range(0, max_depth+1):
        if len(dependency_mapping[depth]) == 0:
            print(depth)
            continue

        output = "\n".join(sorted(dependency_mapping[depth]))

        # add the output to the list of output strings
        outputs.append(output)

    # return the combined output
    return "\n\n".join(outputs)


if __name__ == "__main__":

    # get the command line argument
    import sys
    assert len(sys.argv) == 2, "Usage: python3 dependencies_util.py <dependency tree string>"
    dep_tree = sys.argv[1]

    # generate the ordered list of requirements and print it
    dep_mapping = generate_dependency_mapping(dep_tree)
    requirements_list = generate_dependency_requirements(dep_mapping)
    print(requirements_list)
